/***************************************************************************
 *   kio-locate: KDE I/O Slave for the locate command                      *
 *                                                                         *
 *   Copyright (C) 2004 by Armin Straub                                    *
 *   linux@arminstraub.de                                                  *
 *                                                                         *
 *   This program was initially written by Michael Schuerig.               *
 *   Although I have completely rewritten it, most ideas are adopted       *
 *   from his original work.                                               *
 *                                                                         *
 *   Copyright (C) 2002 by Michael Schuerig                                *
 *   michael@schuerig.de                                                   *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <algorithm>
#include <stdlib.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>

#include <kapplication.h>
#include <kconfigdialog.h>
#include <kdebug.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kurl.h>
#include <kuser.h>
#include <qfile.h>

#include "kio_locate.h"
#include "klocateconfig.h"
#include "klocateconfigwidget.h"
#include "klocateconfiglocatewidget.h"

using namespace KIO;


static const QString queryQuery = "q";
static const QString queryDirectory = "directory";
static const QString queryCase = "case";


/////////////////////////////////////////////////////////////////////
// HELPERS

/**
 * Determines if path includes a trailing slash.
 * @param path the path to inspect
 */
bool hasTrailingSlash(const QString& path)
{
    int n = path.length();
    return ((n > 0) && (path[n-1] == '/'));
}

/**
 * Strips a trailing slash / from a path.
 * @param path the path to strip the slash off
 */
QString stripTrailingSlash(const QString& path)
{
    int n = path.length();
    if ((n > 0) && (path[n-1] == '/')) {
        return path.left(n-1);
    }
    return path;
}

/**
 * Add a trailing slash / to a path if there is none yet.
 * @param path the path to append the slash to
 */
QString addTrailingSlash(const QString& path)
{
    int n = path.length();
    if ((n > 0) && (path[n-1] == '/')) {
        return path;
    }
    return path + '/';
}


/**
 * Count the number of occurences of character c in the string s.
 * @param s the string to search in
 * @param c the character to search for
 */
int countOccurences(const QString& s, QChar c)
{
    int n = 0;
    for (unsigned int i = 0; i < s.length(); i++) {
        if (s[i] == c) {
            n++;
        }
    }
    return n;
}

void addAtom(UDSEntry& entry, unsigned int uds, const QString& s)
{
    UDSAtom a;
    a.m_uds = uds;
    a.m_str = s;
    entry.append(a);
}


void addAtom(UDSEntry& entry, unsigned int uds, long l)
{
    UDSAtom a;
    a.m_uds = uds;
    a.m_long = l;
    entry.append(a);
}


const UDSEntry pathToUDSEntry(const QString& path, const QString& display,
    const QString& url = QString::null, const QString& icon = QString::null)
{
    UDSEntry entry;
    addAtom(entry, KIO::UDS_NAME, display);

    if (!path.isEmpty()) {
        struct stat info;
        lstat(path, &info);

        addAtom(entry, KIO::UDS_SIZE, info.st_size);
        addAtom(entry, KIO::UDS_ACCESS, info.st_mode);
        addAtom(entry, KIO::UDS_MODIFICATION_TIME, info.st_mtime);
        addAtom(entry, KIO::UDS_ACCESS_TIME, info.st_atime);
        addAtom(entry, KIO::UDS_CREATION_TIME, info.st_ctime);

        struct passwd * user = getpwuid(info.st_uid);
        struct group * group = getgrgid(info.st_gid);
        addAtom(entry, KIO::UDS_USER, ((user != NULL) ? user->pw_name : "???" ));
        addAtom(entry, KIO::UDS_GROUP, ((group != NULL) ? group->gr_name : "???" ));

        if (url.isEmpty()) {
            // List an existing file.
            addAtom(entry, KIO::UDS_URL, "file:" + path);

            mode_t type = info.st_mode;
            if (S_ISLNK(type)) {
                QString slink = QString::null;
                char buff[1000];
                int n = readlink(path, buff, 1000);
                if (n != -1) {
                    buff[n] = 0;
                    slink = buff;
                }
                addAtom(entry, KIO::UDS_LINK_DEST, slink);
            } else {
                type &= S_IFMT;
            }
            addAtom(entry, KIO::UDS_FILE_TYPE, type);
        } else {
            // List a locate link.
            addAtom(entry, KIO::UDS_URL, url);
            addAtom(entry, KIO::UDS_FILE_TYPE, S_IFDIR);
        }
    } else {
        addAtom(entry, KIO::UDS_URL, url);
    }

    if (!icon.isEmpty()) {
        addAtom(entry, KIO::UDS_ICON_NAME, icon);
    }

    return entry;
}


/////////////////////////////////////////////////////////////////////
// INITIALIZATION

LocateProtocol::LocateProtocol(const QCString &pool_socket, const QCString &app_socket)
    : SlaveBase("kio_locate", pool_socket, app_socket)
{
    kdDebug() << "LocateProtocol::LocateProtocol()" << endl;

    connect(&m_locater, SIGNAL(found(const QStringList&)),
            this, SLOT(processLocateOutput(const QStringList&)));
    connect(&m_locater, SIGNAL(finished()),
            this, SLOT(locateFinished()));

    m_baseDir = NULL;
    m_curDir = NULL;
}


LocateProtocol::~LocateProtocol()
{
    kdDebug() << "LocateProtocol::~LocateProtocol()" << endl;

    if (m_baseDir != NULL) {
        delete m_baseDir;
    }
}


/////////////////////////////////////////////////////////////////////
// KIO STUFF

void LocateProtocol::setUrl(const KURL& url)
{
    if (url.protocol() != "locater") {
        QString pattern = KURL::decode_string(url.url());
        pattern = pattern.mid(url.protocol().length() + 1);

        KURL newUrl;
        newUrl.setProtocol("locater");

        if (pattern.isEmpty() || pattern == "/") {
            // Give help.
            newUrl.setPath("help");
        } else if (hasTrailingSlash(pattern)) {
            // Detect auto-completion from within konqueror and "stop"
            // this search.
            newUrl.setPath("autosearch");
            newUrl.addQueryItem(queryQuery, pattern);
        } else {
            // Standard plain search.
            newUrl.setPath("search");
            newUrl.addQueryItem(queryQuery, pattern);
        }
        m_url = newUrl;

        kdDebug() << "Redirect: " << m_url << endl;
    } else {
        m_url = url;
    }
}

void LocateProtocol::get(const KURL& url)
{
    kdDebug() << "LocateProtocol::get(" << url << ")" << endl;

    setUrl(url);

    if (isSearchRequest()) {
        error(KIO::ERR_IS_DIRECTORY, QString::null);
    } else if (isConfigRequest()) {
        configRequest();
    } else if (isHelpRequest()) {
        helpRequest();
    } else {
        // What's this?
        error(KIO::ERR_DOES_NOT_EXIST, QString::null);
    }
}


void LocateProtocol::stat(const KURL& url)
{
    kdDebug() << "LocateProtocol::stat(" << url << ")" << endl ;

    setUrl(url);

    if (isSearchRequest()) {
        /// \todo Is UDS_NAME used for anything in stat? If so we should
        /// at least strip of the protocol part.
        UDSEntry entry;
        addAtom(entry, KIO::UDS_NAME, url.decode_string(url.url()));
        addAtom(entry, KIO::UDS_FILE_TYPE, S_IFDIR);
        statEntry(entry);
        finished();
    } else if (isConfigRequest() || isHelpRequest()) {
        UDSEntry entry;
        addAtom(entry, KIO::UDS_NAME, url.decode_string(url.url()));
        addAtom(entry, KIO::UDS_FILE_TYPE, S_IFREG);
        statEntry(entry);
        finished();
        /// \todo Somehow locate: and locate:/ is thought to be a directory
        /// by konqueror anyway. How to change this?
    } else {
        // What's this?
        error(KIO::ERR_DOES_NOT_EXIST, QString::null);
    }
}


void LocateProtocol::listDir(const KURL& url)
{
    kdDebug() << "LocateProtocol::listDir(" << url << ")" << endl ;

    setUrl(url);

    if (isSearchRequest()) {
        searchRequest();
    } else if (isConfigRequest() || isHelpRequest()) {
        error(KIO::ERR_IS_FILE, QString::null);
    } else {
        // What's this?
        error(KIO::ERR_DOES_NOT_EXIST, QString::null);
    }
}


void LocateProtocol::mimetype(const KURL& url)
{
    kdDebug() << "LocateProtocol::mimetype(" << url << ")" << endl ;

    setUrl(url);

    if (isSearchRequest()) {
        mimeType("inode/directory");
    } else if (isConfigRequest() || isHelpRequest()) {
        mimeType("text/html");
    }
    finished();
}


/////////////////////////////////////////////////////////////////////
// SEARCHING

bool LocateProtocol::isSearchRequest()
{
    return m_url.path() == "search";
}

void LocateProtocol::searchRequest()
{
    // Reset old values.
    m_caseSensitivity = caseAuto;
    m_locatePattern = QString::null;
    m_locateDirectory = QString::null;
    m_regExps.clear();

    updateConfig();

    QString query = m_url.queryItem(queryQuery);
    m_locateDirectory = addTrailingSlash(m_url.queryItem(queryDirectory));

    QString caseSensitivity = m_url.queryItem(queryCase);
    if (caseSensitivity == "sensitive") {
        m_caseSensitivity = caseSensitive;
    } else if (caseSensitivity == "insensitive") {
        m_caseSensitivity = caseInsensitive;
    }

    // Split query into components. The first component is the query
    // for locate. The others are filtering regular expressions. They are
    // delimited by (not escaped) whitespace.
    // If the last component starts with two backslahes \\, then the search
    // is only to be done within the directory following them.
    query = query.simplifyWhiteSpace();
    int s = 0;
    int n = query.length();
    for (int i = 0; i <= n; i++) {
        if ((i == n) || ((query[i] == ' ') && (i > 0)
                    && (query[i-1] != '\\') && (i-s > 0))) {
            QString part = partToPattern(query.mid(s, i-s), s==0);
            if (s == 0) {
                m_locatePattern = part;
            } else {
                // For each regular expression determine if it should be
                // case sensitive.
                m_regExps += LocateRegExp(part, !isCaseSensitive(part));
            }
            s = i+1;
        }
    }

    kdDebug() << "Pattern: " << m_locatePattern << endl;

    // Now perform the search...
    infoMessage(i18n("Locating %1 ...").arg(m_locatePattern));

    if (m_baseDir != NULL) {
        delete m_baseDir;
        m_baseDir = NULL;
    }
    m_curDir = NULL;

    m_locater.locate(m_locatePattern, !isCaseSensitive(m_locatePattern));
}


bool LocateProtocol::isCaseSensitive(const QString& text)
{
    if (m_caseSensitivity == caseSensitive) {
        return true;
    } else if (m_caseSensitivity == caseInsensitive) {
        return false;
    } else {
        return text == text.lower();
    }
}


void LocateProtocol::addHit(const QString& path, int subItems)
{
    // kdDebug() << "LocateProtocol::addHit( " << path << ", " << subItems << " )" << endl;

    UDSEntryList entries;

    if (QFile::exists(path)) {
        if (subItems > 0) {
            // Add a directory search.
            entries += pathToUDSEntry(path, pathToDisplay(path, subItems),
                    makeLocaterUrl(path));
        } else {
            // Add a standard file.
            entries += pathToUDSEntry(path, pathToDisplay(path));
        }
    }
    
    listEntries(entries);
}

void LocateProtocol::addPreviousLocateOutput()
{
    if (m_baseDir == NULL) {
        return;
    }
    m_baseDir->prepareListing();
    m_baseDir->listItems(this);
    delete m_baseDir;
    m_baseDir = NULL;
    m_curDir = NULL;
}

bool LocateProtocol::extractBasePath(const QString& path, QString& base)
{
    int n = path.length();
    int c = 0;
    int i = 0;
    if (!m_locateDirectory.isEmpty() && path.startsWith(m_locateDirectory)) {
        i = m_locateDirectory.length();
        c = std::min(countOccurences(m_locateDirectory, '/'), 1);
    }
    while ((i < n) && (c <= m_dontCollapseRootsThreshold)) {
        if (path[i] == '/') {
            c++;
        }
        i++;
    }
    base = path.left(i);
    // if e.g. m_dontCollapseRootsThreshold == 2 then we consider /usr/file to be too short,
    // but /usr/share/file is o.k.
    return (c > m_dontCollapseRootsThreshold); 
}

void LocateProtocol::processLocateOutput(const QStringList& items)
{
    // I don't know if this really necessary, but if we were signaled, we'll
    // better stop.
    if (wasKilled()) {
        m_locater.stop();
        return;
    }
    // Go through what we have found.
    for (QStringList::ConstIterator it = items.begin(); it != items.end(); ++it) {
        QString path = *it;
        if (isMatching(path)) {
            // kdDebug() << "Match: " << path << endl;

            QString base;
            if (!extractBasePath(path, base)) {
                // The path is too short to be put into a directory.
                addHit(path);
            } else {
                if ((m_baseDir != NULL) && !path.startsWith(m_baseDir->m_path)) {
                    addPreviousLocateOutput();
                }
                // Add path to current directory.
                if (m_baseDir == NULL) {
                    m_baseDir = new LocateDirectory(NULL, base, m_collapseDirectoryThreshold);
                    m_curDir = m_baseDir;
                }
                m_curDir = m_curDir->addPath(path);
            }
        }
    }
}

void LocateProtocol::locateFinished()
{
    kdDebug() << "LocateProtocol::locateFinished" << endl;

    // Add any pending items.
    addPreviousLocateOutput();
    
    infoMessage(i18n("Finished."));
    finished();
}


QString LocateProtocol::partToPattern(const QString& part, bool forLocate)
{
    // kdDebug() << "BEG part: " << part << endl;
    QString pattern = part;
    // Unescape whitespace.
    pattern.replace("\\ ", " ");
    // Unquote quoted pattern.
    int n = pattern.length();
    if ((n > 1) && (pattern[0] == '"') && (pattern[n-1] == '"')) {
        pattern = pattern.mid(1, n-2);
    }

    // Special treatment for the pattern used for locate:
    if (forLocate) {
        // Replace + by *. but \+ with +
        int index = -1;
        while ((index = pattern.find('+', index+1)) >= 0) {
            if ((index == 0) || (pattern[index-1] != '\\')) {
                pattern[index] = '*';
            }
        }
        pattern.replace("\\+", "+");

        // Replace ~/ and ~user/ at the beginning (as the shell does)
        if ((pattern.length() > 0) && (pattern[0] == '~')) {
            index = pattern.find('/');
            if (index >= 0) {
                QString name = pattern.mid(1, index-1);
                QString homeDir;
                if (name.isEmpty()) {
                    homeDir = KUser(KUser::UseRealUserID).homeDir();
                } else {
                    homeDir = KUser(name).homeDir();
                }
                if (!homeDir.isEmpty()) {
                    pattern.replace(0, index, homeDir);
                }
            }
        }
        pattern.replace("\\~", "~");
    }
    // kdDebug() << "END part: " << pattern << endl;
    return pattern;
}


bool LocateProtocol::isMatching(const QString& file)
{
    bool matches = true;
    // The file has to belong to our directory.
    matches = file.startsWith(m_locateDirectory);
    // And it has to match against all regular expressions.
    LocateRegExpList::Iterator it = m_regExps.begin();
    for (; matches && (it != m_regExps.end()); ++it) {
        matches = (*it).isMatching(file);
    }
    return matches;
}

QString LocateProtocol::pathToDisplay(const QString& path, int subItems)
{
    // Split off the directory part. If it is not just the minimal '/'.
    QString display = path;
    if ((m_locateDirectory != "/") && display.startsWith(m_locateDirectory)) {
        display = display.mid(m_locateDirectory.length());
    }
    if (subItems > 0) {
        return m_collapsedDisplay.arg(subItems).arg(display);
    }
    return display;
}

QString LocateProtocol::makeLocaterUrl(const QString& directory)
{
    KURL url(m_url);
    url.removeQueryItem(queryDirectory);
    url.addQueryItem(queryDirectory, directory);
    return url.url();
}


/////////////////////////////////////////////////////////////////////
// CONFIG

bool LocateProtocol::isConfigRequest()
{
    return m_url.path() == "config";
}

void LocateProtocol::configRequest()
{
    // Don't show it twice. During my tests I never got there however.
    if(KConfigDialog::showDialog("settings"))
        return;

    KConfigDialog *dialog = new KConfigDialog(0, "settings", KLocateConfig::self(),
            KDialogBase::IconList,
            KDialogBase::Default|KDialogBase::Ok|KDialogBase::Cancel|KDialogBase::Help,
            KDialogBase::Ok, true);
    dialog->setCaption(i18n("Configure - kio-locate"));
    dialog->setIcon(SmallIcon("find"));

    dialog->addPage(new KLocateConfigWidget(), i18n("General"), "package_settings");
    dialog->addPage(new KLocateConfigLocateWidget(), i18n("Locate"), "find");

    // React on user's actions.
    connect(dialog, SIGNAL(settingsChanged()), this, SLOT(updateConfig()));
    connect(dialog, SIGNAL(finished()), this, SLOT(configFinished()));

    dialog->show();
    qApp->enter_loop();
    delete dialog;
}

void LocateProtocol::configFinished()
{
    kdDebug() << "LocateProtocol::configFinished" << endl;

    qApp->exit_loop();
    mimeType("text/html");
    data(QByteArray());
    finished();
}

void LocateProtocol::updateConfig()
{
    kdDebug() << "LocateProtocol::updateConfig" << endl;

    KLocateConfig::self()->readConfig();
    m_collapseDirectoryThreshold = KLocateConfig::collapseDirectoryThreshold();
    m_dontCollapseRootsThreshold = KLocateConfig::dontCollapseRootsThreshold();
    m_collapsedDisplay = KLocateConfig::collapsedDisplay();

    m_locater.setupLocate(KLocateConfig::locateBinary(),
            KLocateConfig::locateAdditionalArguments());
}


/////////////////////////////////////////////////////////////////////
// HELP

bool LocateProtocol::isHelpRequest()
{
    return m_url.path() == "help";
}

void LocateProtocol::helpRequest()
{
    mimeType("text/html");
    QByteArray theData;
    theData = QString( "<html><body><h1>" ).append( i18n("Config") ).append( "</h1></body></html>" ).local8Bit();
    data( theData );
    data(QByteArray());
    finished();
}


/////////////////////////////////////////////////////////////////////
// SEARCH STRUCTURES

LocateDirectory::LocateDirectory(LocateDirectory *parent, const QString& path, int maxHits)
{
    m_parent = parent;
    m_path = path;
    m_maxHits = maxHits;

    m_childs = NULL;
    m_itemsCount = 0;
}

LocateDirectory::~LocateDirectory()
{
    if (m_childs != NULL) {
        delete m_childs;
    }
}

LocateDirectory *LocateDirectory::addPath(const QString& path)
{
    // kdDebug() << "LocateDirectory::addPath( " << path << " ) - path: " << m_path << endl;

    if (path.startsWith(m_path)) {
        QString relPath = path.mid(m_path.length());
        int p = relPath.findRev('/');
        if (p >= 0) {
            LocateDirectory *child = getSubDirectory(relPath.left(p));
            child->addItem(relPath.mid(p+1));
            return child;
        }
        addItem(relPath);
        return this;
    }
    if (m_parent != NULL) {
        return m_parent->addPath(path);
    }
    // This should not happen
    return this;
}

LocateDirectory *LocateDirectory::getSubDirectory(const QString& relPath)
{
    // kdDebug() << "LocateDirectory::getSubDirectory( " << relPath << " ) - path: " << m_path << endl;

    if (m_childs == NULL) {
        m_childs = new LocateDirectories();
        m_childs->setAutoDelete(true);
    }
    QString base = relPath;
    int p = relPath.find('/');
    if (p >= 0) {
        base = relPath.left(p);
    }
    LocateDirectory *child = m_childs->find(base);
    if (child == NULL) {
        child = new LocateDirectory(this, m_path + base + '/', m_maxHits);
        m_childs->insert(base, child);
    }
    if (p >= 0) {
        return child->getSubDirectory(relPath.mid(p+1));
    }
    return child;
}

void LocateDirectory::addItem(const QString& path)
{
    // kdDebug() << "LocateDirectory::addItem( " << path << " ) - path: " << m_path << endl;

    if (m_itemsCount < m_maxHits) {
        m_items += LocateItem(m_path + path, 0);
    }
    m_itemsCount++;
}

void LocateDirectory::prepareListing()
{
    // kdDebug() << "LocateDirectory::prepareListing() - path: " << m_path << endl;

    // Set m_fullCount to the total number of childs.
    m_fullCount = m_itemsCount;

    if (m_childs != NULL) {
        LocateDirectoriesIterator child(*m_childs);
        for (; child.current(); ++child) {
            child.current()->prepareListing();
        }
    }
    if (m_parent != NULL) {
        m_parent->m_fullCount += m_fullCount;
        if (m_itemsCount > m_maxHits) {
            m_parent->m_items += LocateItem(m_path, m_fullCount);
            m_parent->m_itemsCount++;
        } else {
            m_parent->m_items += m_items;
            m_parent->m_itemsCount += m_itemsCount;
        }
    } else {
        if (m_itemsCount > m_maxHits) {
            m_items.clear();
            m_items += LocateItem(m_path, m_fullCount);
        }
    }
}

void LocateDirectory::listItems(LocateProtocol *protocol)
{
    // kdDebug() << "LocateDirectory::listItems() - path: " << m_path << endl;
    // kdDebug() << "items: " << m_itemsCount << " - really: " << m_items.count() << endl;

    LocateItems::ConstIterator item = m_items.begin();
    for (; item != m_items.end(); ++item) {
        protocol->addHit((*item).m_path, (*item).m_subItems);
    }
}

LocateItem::LocateItem()
{
}

LocateItem::LocateItem(const QString& path, int subItems)
{
    m_path = path;
    m_subItems = subItems;
}


/////////////////////////////////////////////////////////////////////
// INVOKATION

extern "C"
{
    int kdemain(int argc, char **argv)
    {
        // We use KApplication instead of KInstance here, because we use a
        // config dialog and such gui stuff.
        KApplication app(argc, argv, "kio_locate", false, true);

        kdDebug(7101) << "*** Starting kio_locate " << endl;

        if (argc != 4) {
            kdDebug(7101) << "Usage: kio_locate  protocol domain-socket1 domain-socket2" << endl;
            exit(-1);
        }

        LocateProtocol slave(argv[2], argv[3]);
        slave.dispatchLoop();

        kdDebug(7101) << "*** kio_locate Done" << endl;
        return 0;
    }
}


#include "kio_locate.moc"
